
/*
 * UAE Action Replay 3 and HRTMon support
 *
 * (c) 2000 Toni Wilen <twilen@arabuusimiehet.com>
 *
 * My unofficial UAE patches will be located at
 * http://www.arabuusimiehet.com/~twilen/uae/
 *
 * HRTMon:
 *
 * HRTMon support is tested with version 2.25 + patch.
 * More information about HRTMon can be found from
 * http://dumbo.cryogen.ch/hrtmon/
 * 
 * Action Replay 3:
 *
 * Tested with ROM version 3.09 (10/13/91) (Are there other AR3 ROMs?)
 * This patch also makes AR3 compatible with KickStart's other than 1.3
 * (ROM checksum error is normal with KS != 1.3)  
 * NOTE: AR3 has problems with 68020+ processors.
 * For maximum compatibility select 68000/68010 and A500 speed from UAE
 * options. AR3 disk code is also very bad and does not work reliably
 * without my "cycle-exact" floppy emulation code.
 *
 * How to rip Action Replay 3 ROM:
 *
 * Find A500 with AR3, press 'freeze'-button
 *
 * write following:
 *
 * may<RETURN>
 * the<RETURN>
 * force<RETURN>
 * be<RETURN>
 * with<RETURN>
 * you<RETURN>
 * new<RETURN>
 * 
 * now 256K ROM is visible at 0x400000-0x43ffff
 * and 64K RAM at 0x440000-0x44ffff
 *
 * following command writes ROM to disk:
 *
 * sm ar3.rom,400000 440000
 *
 * I DO NOT REPLY MAILS ASKING FOR ACTION REPLAY ROMS!
 *
 * AR3 hardware notes (not 100% correct..)
 *
 * first 8 bytes of ROM are not really ROM, they are
 * used to read/write cartridge's hardware state
 *
 * 0x400000: hides cartridge ROM/RAM when written
 * 0x400001: read/write HW state
 *  3 = reset (read-only)
 *  2 = sets HW to activate when breakpoint condition is detected
 *  1 = ???
 *  0 = freeze pressed
 * 0x400002/0x400003: mirrors 0x400000/0x400001
 * 0x400006/0x400007: when written to, turns chip-ram overlay off
 *
 * breakpoint condition is detected when CPU first accesses
 * chip memory below 1024 bytes and then reads CIA register
 * $BFE001.
 *
 * cartridge hardware also snoops CPU accesses to custom chip
 * registers (DFF000-DFF1FE). All CPU custom chip accesses are
 * saved to RAM at 0x44f000-0x44f1ff. Note that emulated AR3 also 
 * saves copper's custom chip accesses. This fix stops most AR3
 * detection tricks.
 *
 * Patch tested with WinUAE 0.8.12 beta 1 and UAE 0.8.12 (Linux)
 *
 */

#include "sysconfig.h"
#include "sysdeps.h"

#include "config.h"
#include "options.h"
#include "uae.h"
#include "memory.h"
#include "custom.h"
#include "readcpu.h"
#include "newcpu.h"
#include "zfile.h"
#include "ar.h"

/* HRTMon baseaddress, can be freely changed */
#define HRTMON_BASE 0x980000

uae_u8 ar_custom[2*256];

int hrtmon_flag = ACTION_REPLAY_INACTIVE;

uae_u8 *hrtmemory;
static uae_u32 hrtmem_mask;
static uae_u8 *hrtmon_custom;
uae_u32 hrtmem_start, hrtmem_size;

static uae_u32 hrtmem_lget (uaecptr) REGPARAM;
static uae_u32 hrtmem_wget (uaecptr) REGPARAM;
static uae_u32 hrtmem_bget (uaecptr) REGPARAM;
static void  hrtmem_lput (uaecptr, uae_u32) REGPARAM;
static void  hrtmem_wput (uaecptr, uae_u32) REGPARAM;
static void  hrtmem_bput (uaecptr, uae_u32) REGPARAM;
static int  hrtmem_check (uaecptr addr, uae_u32 size) REGPARAM;
static uae_u8 *hrtmem_xlate (uaecptr addr) REGPARAM;

static uae_u32 REGPARAM2 hrtmem_lget (uaecptr addr)
{
    uae_u32 *m;
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    m = (uae_u32 *)(hrtmemory + addr);
    return do_get_mem_long (m);
}

static uae_u32 REGPARAM2 hrtmem_wget (uaecptr addr)
{
    uae_u16 *m;
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    m = (uae_u16 *)(hrtmemory + addr);
    return do_get_mem_word (m);
}

static uae_u32 REGPARAM2 hrtmem_bget (uaecptr addr)
{
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    return hrtmemory[addr];
}

static void REGPARAM2 hrtmem_lput (uaecptr addr, uae_u32 l)
{
    uae_u32 *m;
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    m = (uae_u32 *)(hrtmemory + addr);
    do_put_mem_long (m, l);
}

static void REGPARAM2 hrtmem_wput (uaecptr addr, uae_u32 w)
{
    uae_u16 *m;
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    m = (uae_u16 *)(hrtmemory + addr);
    do_put_mem_word (m, (uae_u16)w);
}

static void REGPARAM2 hrtmem_bput (uaecptr addr, uae_u32 b)
{
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    hrtmemory[addr] = b;
}

static int REGPARAM2 hrtmem_check (uaecptr addr, uae_u32 size)
{
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    return (addr + size) <= hrtmem_size;
}

static uae_u8 REGPARAM2 *hrtmem_xlate (uaecptr addr)
{
    addr -= hrtmem_start & hrtmem_mask;
    addr &= hrtmem_mask;
    return hrtmemory + addr;
}

#ifdef PATCHED_MEMORY_MANAGEMENT
static struct bankinfo hrtmon_bank = {
    "HRTMon memory",0,
    { hrtmem_lget, hrtmem_wget, hrtmem_bget,
      hrtmem_lput, hrtmem_wput, hrtmem_bput,
      hrtmem_xlate, hrtmem_check }
};
#else
addrbank hrtmem_bank = {
    hrtmem_lget, hrtmem_wget, hrtmem_bget,
    hrtmem_lput, hrtmem_wput, hrtmem_bput,
    hrtmem_xlate, hrtmem_check
};
#endif

static void copyfromamiga(uae_u8 *dst,uae_u8 *src,int len)
{
while(len--) {
    *dst++=get_byte((uaecptr)src);
    src++;
}
}
static void copytoamiga(uae_u8 *dst,uae_u8 *src,int len)
{
while(len--) {
    put_byte((uaecptr)dst,*src++);
    dst++;
}
}

int action_replay_flag = ACTION_REPLAY_INACTIVE;

static uae_u8 artemp[4];
static uae_u8 armode;

static void ar3write(uaecptr addr, uae_u8 b)
{
if(addr>=8) return;
/* write_log("%02.2X written to %08.8X\n",b,addr); */
if(action_replay_flag == ACTION_REPLAY_ACTIVE) {
    if (addr == 1) {
	armode = b;
        if(b == 2) {
	    write_log("AR: exit with breakpoint(s) active\n");
	} else {
	    write_log("AR: exit\n");
	}
	regs.spcflags |= SPCFLAG_ACTION_REPLAY;
	action_replay_flag = ACTION_REPLAY_HIDE;
    } else if(addr == 6) {
	copytoamiga((uae_u8*)regs.vbr+0x7c,artemp,4);
	write_log("AR: chipmem returned\n");
    }
}
}

uae_u8 *armemory;
#define armem_start 0x400000
#define armem_size  0x080000

static uae_u32 armem_lget (uaecptr) REGPARAM;
static uae_u32 armem_wget (uaecptr) REGPARAM;
static uae_u32 armem_bget (uaecptr) REGPARAM;
static void  armem_lput (uaecptr, uae_u32) REGPARAM;
static void  armem_wput (uaecptr, uae_u32) REGPARAM;
static void  armem_bput (uaecptr, uae_u32) REGPARAM;
static int  armem_check (uaecptr addr, uae_u32 size) REGPARAM;
static uae_u8 *armem_xlate (uaecptr addr) REGPARAM;

static uae_u32 REGPARAM2 armem_lget (uaecptr addr)
{
    uae_u32 *m;
    addr -= armem_start;
    m = (uae_u32 *)(armemory + addr);
    return do_get_mem_long (m);
}

static uae_u32 REGPARAM2 armem_wget (uaecptr addr)
{
    uae_u16 *m;
    addr -= armem_start;
    m = (uae_u16 *)(armemory + addr);
    return do_get_mem_word (m);
}

static uae_u32 REGPARAM2 armem_bget (uaecptr addr)
{
    addr -= armem_start;
    return armemory[addr];
}

static void REGPARAM2 armem_lput (uaecptr addr, uae_u32 l)
{
    uae_u32 *m;
    if(addr >= 0x400008 && addr < 0x440000) return;
    addr -= armem_start;
    m = (uae_u32 *)(armemory + addr);
    if(addr >= 0x40000) do_put_mem_long (m, l);
    ar3write(addr+0,(uae_u8)(l>>24));
    ar3write(addr+1,(uae_u8)(l>>16));
    ar3write(addr+2,(uae_u8)(l>>8));
    ar3write(addr+3,(uae_u8)(l>>0));
}

static void REGPARAM2 armem_wput (uaecptr addr, uae_u32 w)
{
    uae_u16 *m;
    if(addr >= 0x400008 && addr < 0x440000) return;
    addr -= armem_start;
    m = (uae_u16 *)(armemory + addr);
    if(addr >= 0x40000) do_put_mem_word (m, (uae_u16)w);
    ar3write(addr+0,(uae_u8)(w>>8));
    ar3write(addr+1,(uae_u8)(w>>0));
}

static void REGPARAM2 armem_bput (uaecptr addr, uae_u32 b)
{
    if(addr >= 0x400008 && addr < 0x440000) return;
    addr -= armem_start;
    if(addr >= 0x40000) armemory[addr] = b;
    ar3write(addr+0,b);
}

static int REGPARAM2 armem_check (uaecptr addr, uae_u32 size)
{
    addr -= armem_start;
    return (addr + size) <= armem_size;
}

static uae_u8 REGPARAM2 *armem_xlate (uaecptr addr)
{
    addr -= armem_start;
    return armemory + addr;
}

#ifdef PATCHED_MEMORY_MANAGEMENT
static struct bankinfo actionreplay_bank = {
    "Action Replay memory",0,
    { armem_lget, armem_wget, armem_bget,
      armem_lput, armem_wput, armem_bput,
      armem_xlate, armem_check }
};
#else
addrbank armem_bank = {
    armem_lget, armem_wget, armem_bget,
    armem_lput, armem_wput, armem_bput,
    armem_xlate, armem_check
};
extern addrbank dummy_bank;
#endif


static void hide_cart(int hide)
{
#ifdef ACTION_REPLAY_HIDE_CARTRIDGE
if(hide) {
    #ifdef PATCHED_MEMORY_MANAGEMENT
    map_banks(BANK_NONE,0x40,8);
    #else
    map_banks(&dummy_bank,0x40,8);
    #endif
} else {	
    #ifdef PATCHED_MEMORY_MANAGEMENT
    map_banks (BANK_ACTIONREPLAY, 0x40, 8);
    #else
    map_banks (&armem_bank, 0x40, 8);
    #endif
}
#endif
}

extern void Interrupt(int nr);

/* Cartridge activates itself by overlaying its rom
 * over chip-ram and then issuing IRQ 7
 *
 * I just copy IRQ vector 7 from ROM to chip RAM
 * instead of fully emulating cartridge's behaviour..
 */

static void action_replay_go(void)
{
hide_cart(0);
memcpy(armemory+262144+0xf000,ar_custom,2*256);
action_replay_flag = ACTION_REPLAY_ACTIVE;
regs.spcflags |= SPCFLAG_ACTION_REPLAY;
copyfromamiga(artemp,(uae_u8*)regs.vbr+0x7c,4);
copytoamiga((uae_u8*)regs.vbr+0x7c,armemory+0x7c,4);
Interrupt(7);
}

static void hrtmon_go(int mode)
{
memcpy(hrtmon_custom,ar_custom,2*256);
hrtmon_flag = ACTION_REPLAY_ACTIVE;
regs.spcflags |= SPCFLAG_ACTION_REPLAY;
put_long((uaecptr)((uae_u8*)regs.vbr+0x7c), hrtmem_start + 8 + (mode ? 4 : 0));
Interrupt(7);
}

void hrtmon_enter(void)
{
if (!hrtmemory) return;
write_log("HRTMON: freeze\n");
hrtmon_go(1);
}

void action_replay_enter(void)
{
if (!armemory) return;
if (action_replay_flag == ACTION_REPLAY_DORESET) {
    write_log("AR: reset\n");
    armemory[1]=3;
    action_replay_go();
    return;
}
switch (armode)
    {
    case 0:
    write_log("AR: activated (freeze)\n");
    armemory[1]=0;
    break;
    case 2:
    write_log("AR: activated (breakpoint)\n");
    armemory[1]=1;
    break;
}
action_replay_go();
}

void action_replay_reset(void)
{
if(action_replay_flag == ACTION_REPLAY_INACTIVE) return;
action_replay_flag = ACTION_REPLAY_WAITRESET;
}

void action_replay_ciaread(void)
{
if (action_replay_flag != ACTION_REPLAY_IDLE) return;
if (action_replay_flag == ACTION_REPLAY_INACTIVE) return;
if (armode != 2) return;
if (m68k_getpc() >= 0x200) return;
action_replay_flag = ACTION_REPLAY_ACTIVATE;
regs.spcflags |= SPCFLAG_ACTION_REPLAY;
}

void action_replay_freeze(void)
{
if(action_replay_flag == ACTION_REPLAY_IDLE) {
    action_replay_flag = ACTION_REPLAY_ACTIVATE;
    regs.spcflags |= SPCFLAG_ACTION_REPLAY;
} else if(hrtmon_flag == ACTION_REPLAY_IDLE) {
    hrtmon_flag = ACTION_REPLAY_ACTIVATE;
    regs.spcflags |= SPCFLAG_ACTION_REPLAY;
}
}

void action_replay_chipwrite(void)
{
    action_replay_flag = ACTION_REPLAY_DORESET;
    regs.spcflags |= SPCFLAG_ACTION_REPLAY;
}

void action_replay_hide(void)
{
hide_cart(1);
action_replay_flag = ACTION_REPLAY_IDLE;
}

void hrtmon_hide(void)
{
hrtmon_flag = ACTION_REPLAY_IDLE;
write_log("HRTMON: exit\n");
}

void hrtmon_breakenter(void)
{
hrtmon_flag = ACTION_REPLAY_HIDE;
regs.spcflags |= SPCFLAG_ACTION_REPLAY;
write_log("HRTMON: breakpoint\n");
}

/* Original AR3 only works with KS 1.3
 * this patch fixes that problem.
 */

static uae_u8 arpatch1[]={0x20,0xc9,0x51,0xc9,0xff,0xfc};
static uae_u8 arpatch2[]={0x00,0xfc,0x01,0x44};

#ifndef PATCHED_MEMORY_MANAGEMENT
extern uae_u8 *kickmemory;
#endif

static void action_replay_patch(void)
{
int off1,off2;
#ifdef PATCHED_MEMORY_MANAGEMENT
uae_u8 *kickmem = memorybanks[BANK_KICK].addr;
#else
uae_u8 *kickmem = kickmemory;
#endif

return;
if(!memcmp(kickmem,kickmem+262144,262144)) off1=262144; else off1=0;
for(;;) {  
    if(!memcmp(kickmem+off1,arpatch1,sizeof(arpatch1))) break;
    off1++;
}
off2=0;
for(;;) {
    if(!memcmp(armemory+off2,arpatch2,sizeof(arpatch2))) break;
    off2++;
}
armemory[off2+0]=(uae_u8)((off1+kickmem_start+2)>>24);
armemory[off2+1]=(uae_u8)((off1+kickmem_start+2)>>16);
armemory[off2+2]=(uae_u8)((off1+kickmem_start+2)>>8);
armemory[off2+3]=(uae_u8)((off1+kickmem_start+2)>>0);
}

int action_replay_load(int activate)
{
FILE *f;

action_replay_flag = ACTION_REPLAY_INACTIVE;
#ifdef PATCHED_MEMORY_MANAGEMENT
create_memorybank (BANK_ACTIONREPLAY, &actionreplay_bank);
alloc_bank(BANK_ACTIONREPLAY,262144+65536,262144+65536);
armemory = memorybanks[BANK_ACTIONREPLAY].addr;
#else
armemory = xmalloc (262144+65536);
#endif
memset(armemory,0,262144+65536);
f = zfile_open("ar3.rom","rb");
if (!f) {
    write_log("failed to load Action Replay ROM\n");
    return 0;
}
fread (armemory,262144,1,f);
fclose (f);
#ifdef PATCHED_MEMORY_MANAGEMENT
map_banks (BANK_ACTIONREPLAY, 0x40, 8);
#else
map_banks (&armem_bank, 0x40, 8);
#endif
action_replay_patch();
hide_cart(1);
if(activate) action_replay_flag = ACTION_REPLAY_WAITRESET;
write_log("Action Replay installed at %08.8X, size %08.8X\n",armem_start,armem_size);
return 1;
}

#ifndef FALSE
#define FALSE 0
#endif
#ifndef TRUE
#define TRUE 1
#endif

typedef struct {
	char jmps[20];
	unsigned int mon_size;
	unsigned short col0, col1;
	char right;
	char keyboard;
	char key;
	char ide;
	char a1200;
	char aga;
	char insert;
	char delay;
	char lview;
	char cd32;
	char screenmode;
	char vbr;
	char entered;
	char hexmode;
	unsigned short error_sr;
	unsigned int error_pc;
	unsigned short error_status;
	char newid[6];
	unsigned short	mon_version;
	unsigned short	mon_revision;
	unsigned int	whd_base;
	unsigned short	whd_version;
	unsigned short	whd_revision;
	unsigned int	max_chip;
	unsigned int	custom;
} HRTCFG;

static void hrtmon_configure(HRTCFG *cfg)
{
do_put_mem_word(&cfg->col0,0x000);
do_put_mem_word(&cfg->col1,0xeee);
cfg->right = 0;
cfg->key = FALSE;
cfg->ide = 0;
cfg->a1200 = 0;
cfg->aga = (currprefs.chipset_mask & 4) ? 1 : 0;
cfg->insert = TRUE;
cfg->delay = 15;
cfg->lview = FALSE;
cfg->cd32 = 0;
cfg->screenmode = 0;
cfg->vbr = TRUE;
cfg->hexmode = FALSE;
cfg->mon_size=0;
cfg->hexmode = TRUE;
do_put_mem_long(&cfg->max_chip,currprefs.chipmem_size);
hrtmon_custom = do_get_mem_long ((uae_u32*)&cfg->custom)+hrtmemory;
}

static void hrtmon_reloc(uae_u32 *mem,uae_u32 *header)
{
uae_u32 *p;
uae_u8 *base;
uae_u32 len, i;

len=do_get_mem_long(header+7);
p=mem+len+3;
len=do_get_mem_long(p-2);
for(i=0;i<len;i++) {
    base=(uae_u8*)do_get_mem_long(p)+(uae_u32)mem;
    do_put_mem_long((uae_u32*)base, do_get_mem_long((uae_u32*)base)+hrtmem_start);
    p++;
}
}

int hrtmon_load(int activate)
{
FILE *f;
int size;
uae_u32 header[8];

f=zfile_open("hrtmon.rom","rb");
if(!f) {
    write_log("failed to load HRTMon ROM\n");
    return 0;
}
fseek(f,0,SEEK_END);
size = ftell(f) - 8*4;
fseek(f,0,SEEK_SET);
fread(header,sizeof(header),1,f);
hrtmem_size = size;
hrtmem_size += 65535;
hrtmem_size &= ~65535;
hrtmem_mask = hrtmem_size - 1;
hrtmem_start = HRTMON_BASE;
#ifdef PATCHED_MEMORY_MANAGEMENT
create_memorybank(BANK_ACTIONREPLAY, &hrtmon_bank);
alloc_bank(BANK_ACTIONREPLAY,hrtmem_size,hrtmem_size);
hrtmemory = memorybanks[BANK_ACTIONREPLAY].addr;
#else
hrtmemory = xmalloc (hrtmem_size);
#endif
fread (hrtmemory,size,1,f);
fclose (f);
hrtmon_configure((HRTCFG*)hrtmemory);
hrtmon_reloc((uae_u32*)hrtmemory,header);
#ifdef PATCHED_MEMORY_MANAGEMENT
map_banks (BANK_ACTIONREPLAY, hrtmem_start >> 16, hrtmem_size >> 16);
#else
map_banks (&hrtmem_bank, hrtmem_start >> 16, hrtmem_size >> 16);
#endif
hrtmon_flag = ACTION_REPLAY_IDLE;
if(!activate) hrtmon_flag = ACTION_REPLAY_INACTIVE;
write_log("HRTMon installed at %08.8X, size %08.8X\n",hrtmem_start >> 16, hrtmem_size >> 16);
return 1;
}
